//==========================================================================
//
//      bplist-dynamic.c
//
//      Dynamic breakpoint list.
//      Currently only statically allocated.  (ie NO_MALLOC is assumed)
//
//==========================================================================

//==========================================================================
//
// Purpose:      Dynamic breakpoint list.
// Description:  
//               
//=========================================================================

#include <include/pkgconf/system.h>
#include <include/pkgconf/hal.h>

#if defined(RTNUM_HAL_BREAKPOINT_LIST_SIZE) && (RTNUM_HAL_BREAKPOINT_LIST_SIZE > 0) && defined(RTDBG_HAL_DEBUG_GDB_INCLUDE_STUBS)

#define RTARC_HAL_COMMON_EXPORT_CPU_MACROS
#include <include/hal/hal_stub.h>
#include <include/hal/hal_arch.h>
#include <include/hal/hal_cache.h>

/*
 * A simple target breakpoint list without using malloc.
 * To use this package, you must define HAL_BREAKINST_SIZE to be the size
 * in bytes of a trap instruction (max if there's more than one),
 * HAL_BREAKINST to the opcode value of the instruction, and
 * HAL_BREAKINST_TYPE to be the type necessary to hold the opcode value.
 */

struct breakpoint_list {
  target_register_t  addr;
  char old_contents [HAL_BREAKINST_SIZE];
  struct breakpoint_list *next;
  char in_memory;
  char length;
} *breakpoint_list = NULL;

#ifndef HAL_BREAKINST_ADDR
static HAL_BREAKINST_TYPE break_inst = HAL_BREAKINST;
#define HAL_BREAKINST_ADDR(x) ((void*)&break_inst)
#endif

/*	RTNUM_HAL_BREAKPOINT_LIST_SIZE will be defined latter by Platform Builder 
	in the exported header file <pkgconf/hal.h>.							*/
static struct breakpoint_list bp_list [RTNUM_HAL_BREAKPOINT_LIST_SIZE];
static struct breakpoint_list *free_bp_list = NULL;
static int curr_bp_num = 0;


/*	This is the function that will be called by GDB stub when the GDB user
**	set a breakpoint on his/her host PC using GDB debugger.
**	addr:	is the address of the breakpoint-victim instruction
**	len:	...
*/
int
__set_breakpoint (target_register_t addr, target_register_t len)
{
	struct breakpoint_list **add_ent = &breakpoint_list;
	struct breakpoint_list *l = breakpoint_list;
	struct breakpoint_list *new_ent;

	while (l != NULL && l->addr < addr)
	{
		add_ent = &l->next;
		l =  l->next;
	}

	if (l != NULL && l->addr == addr)
		return 2;

	if (free_bp_list != NULL)
	{
		new_ent = free_bp_list;
		free_bp_list = free_bp_list->next;
	}
	else
	{
		if (curr_bp_num < RTNUM_HAL_BREAKPOINT_LIST_SIZE)
		{
			new_ent = &bp_list[curr_bp_num++];
		}
		else
		{
			return 1;
		}
	}

	new_ent->addr = addr;
	new_ent->in_memory = 0;
	new_ent->next = l;
	new_ent->length = len;
	*add_ent = new_ent;

	return 0;
}

/*	This is the function that will be called when the GDB user unmark a breakpoint
**	on his/her host PC using GDB debugger.
**	addr:	is the address of the breakpoint-victim instruction
**	len:	...
*/
int
__remove_breakpoint (target_register_t addr, target_register_t len)
{
	struct breakpoint_list *l = breakpoint_list;
	struct breakpoint_list *prev = NULL;

	while (l != NULL && l->addr < addr)
	{
		prev = l;
		l = l->next;
	}

	if ((l == NULL) || (l->addr != addr))
		return 1;

	if (l->in_memory)
	{
		__write_mem_safe (	&l->old_contents[0],
							(void*)l->addr,
							sizeof (l->old_contents)	);
	}

	if (prev == NULL)
		breakpoint_list = l->next;
	else
		prev->next = l->next;

	l->next = free_bp_list;
	free_bp_list = l;

	return 0;
}

/*	All breakpoints(instruction "break 5") are put into memory occupied
**	by the executing image. The "breakpoint victim" instructions are first saved
**	before being overwritten by the "break 5" instructions.					*/
void
__install_breakpoint_list (void)
{
	struct breakpoint_list *l = breakpoint_list;

	while (l != NULL)
	{
		if (! l->in_memory)
		{
			int len = sizeof (l->old_contents);
			if (__read_mem_safe (&l->old_contents[0], (void*)l->addr, len) == len)
			{
				if (__write_mem_safe (HAL_BREAKINST_ADDR(l->length),
								(void*)l->addr, l->length) == l->length)
				{
					l->in_memory = 1;
				}
			}
		}
		l = l->next;
    }
	HAL_ICACHE_SYNC();
}

/*	Restore the previous saved "breakpoint victim" instructions.
**	Breakpoint instructions (like "break 5") will be overwritten by the original ones. */
void
__clear_breakpoint_list (void)
{
	struct breakpoint_list *l = breakpoint_list;

	while (l != NULL)
    {
		if (l->in_memory)
		{
			int len = sizeof (l->old_contents);
			if (__write_mem_safe (&l->old_contents[0], (void*)l->addr, len) == len)
			{
				l->in_memory = 0;
			}
		}
		l = l->next;
    }
	HAL_ICACHE_INVALIDATE_ALL();
}


int
__display_breakpoint_list (void (*print_func)(target_register_t))
{
  struct breakpoint_list *l = breakpoint_list;

  while (l != NULL)
    {
      print_func(l->addr);
      l = l->next;
    }

  return 0;
}

#else  // (RTNUM_HAL_BREAKPOINT_LIST_SIZE == 0) or UNDEFINED

#include <rt/hal/hal_stub.h>           // Our header

#ifndef RTDBG_HAL_DEBUG_GDB_INCLUDE_STUBS
// We don't know that type target_register_t is yet.
// Let's just pick a type so we can compile.  Since
// these versions of the functions don't actually do
// anything with the parameters, the actualy types
// don't matter.
typedef unsigned long target_register_t;
#endif

int
__set_breakpoint (target_register_t addr, target_register_t len)
{
  return 1;
}

int
__remove_breakpoint (target_register_t addr, target_register_t len)
{
  return 1;
}

void
__install_breakpoint_list (void)
{
}

void
__clear_breakpoint_list (void)
{
}

int
__display_breakpoint_list (void (*print_func)(target_register_t))
{
    return 0;
}
#endif // (RTNUM_HAL_BREAKPOINT_LIST_SIZE > 0)

