#ifndef RTONCE_INFRA_RT_TYPE_H
#define RTONCE_INFRA_RT_TYPE_H

//==========================================================================
//
//      rt_type.h
//
//      Standard types, and some useful coding macros.
//
//==========================================================================
//
// Purpose:     share unambiguously sized types.
// Description: we typedef [rt_][u]int8,16,32 &c for general use.
// Usage:       #include "rt/infra/rt_type.h"
//
//==========================================================================

#include <stddef.h>           // Definition of NULL from the compiler

// -------------------------------------------------------------------------
// Some useful macros. These are defined here by default.

// __externC is used in mixed C/C++ headers to force C linkage on an external
// definition. It avoids having to put all sorts of ifdefs in.

#ifdef __cplusplus
# define __externC extern "C"
#else
# define __externC extern
#endif
// Also define externC for now - but it is deprecated
#define externC __externC


//-------------------------------------------------------------------------
//                   Cross Compiler Definitions
//-------------------------------------------------------------------------
#define	COMPILER_GNU_GCC		1
#define	COMPILER_GREEN_HILLS	2

// -------------------------------------------------------------------------
// The header <basetype.h> defines the base types used here. It is
// supplied either by the target architecture HAL, or by the host
// porting kit. They are all defined as macros, and only those that
// make choices other than the defaults given below need be defined.

#define RT_LSBFIRST 1234
#define RT_MSBFIRST 4321

/*
#include <include\hal\basetype.h>
*/
#include "../hal/basetype.h"
#if (RT_BYTEORDER != RT_LSBFIRST) && (RT_BYTEORDER != RT_MSBFIRST)
# error You must define RT_BYTEORDER to equal RT_LSBFIRST or RT_MSBFIRST
#endif

#ifndef RT_DOUBLE_BYTEORDER
#define RT_DOUBLE_BYTEORDER RT_BYTEORDER
#endif

#ifndef rt_halint8
# define rt_halint8 char
#endif
#ifndef rt_halint16
# define rt_halint16 short
#endif
#ifndef rt_halint32
# define rt_halint32 int
#endif
#ifndef rt_halint64
# define rt_halint64 long long
#endif

#ifndef rt_halcount8
# define rt_halcount8 int
#endif
#ifndef rt_halcount16
# define rt_halcount16 int
#endif
#ifndef rt_halcount32
# define rt_halcount32 int
#endif
#ifndef rt_halcount64
# define rt_halcount64 long long
#endif

#ifndef rt_haladdress
# define rt_haladdress rt_uint32
#endif
#ifndef rt_haladdrword
# define rt_haladdrword rt_uint32
#endif

#ifndef rt_halbool
# define rt_halbool int
#endif

#ifndef rt_halatomic
# define rt_halatomic rt_halint8
#endif

// -------------------------------------------------------------------------
// Provide a default architecture alignment
// This may be overridden in basetype.h if necessary.
// These should be straightforward numbers to allow use in assembly.

#ifndef RTARC_ALIGNMENT
# define RTARC_ALIGNMENT 8
#endif
// And corresponding power of two alignment
#ifndef RTARC_P2ALIGNMENT
# define RTARC_P2ALIGNMENT 3
#endif
#if (RTARC_ALIGNMENT) != (1 << RTARC_P2ALIGNMENT)
# error "Inconsistent RTARC_ALIGNMENT and RTARC_P2ALIGNMENT values"
#endif

// -------------------------------------------------------------------------
// The obvious few that compilers may define for you.
// But in case they don't:

#ifndef NULL
# define NULL 0
#endif

#ifndef __cplusplus

typedef rt_halbool bool;

# ifndef false
#  define false 0
# endif

# ifndef true
#  define true (!false)
# endif

#endif

// -------------------------------------------------------------------------
// Allow creation of procedure-like macros that are a single statement,
// and must be followed by a semi-colon

#define RT_MACRO_START do {
#define RT_MACRO_END   } while (0)

#define RT_EMPTY_STATEMENT RT_MACRO_START RT_MACRO_END

#define RT_UNUSED_PARAM( _type_, _name_ ) RT_MACRO_START      \
  _type_ __tmp1 = (_name_);                                     \
  _type_ __tmp2 = __tmp1;                                       \
  __tmp1 = __tmp2;                                              \
RT_MACRO_END


// -------------------------------------------------------------------------
// Reference a symbol without explicitly making use of it. Ensures that
// the object containing the symbol will be included when linking.

#define RT_REFERENCE_OBJECT(__object__)                                 \
     RT_MACRO_START                                                     \
     static void *__rtvar_discard_me__ __attribute__ ((unused)) =       \
                                                          &(__object__); \
     RT_MACRO_END

// -------------------------------------------------------------------------
// Define basic types for using integers in memory and structures;
// depends on compiler defaults and CPU type.

typedef unsigned rt_halint8    rt_uint8  ;
typedef   signed rt_halint8    rt_int8   ;

typedef unsigned rt_halint16   rt_uint16 ;
typedef   signed rt_halint16   rt_int16  ;

typedef unsigned rt_halint32   rt_uint32 ;
typedef   signed rt_halint32   rt_int32  ;

typedef unsigned rt_halint64   rt_uint64 ;
typedef   signed rt_halint64   rt_int64  ;

typedef  rt_halbool            rt_bool   ;

// -------------------------------------------------------------------------
// Define types for using integers in registers for looping and the like;
// depends on CPU type, choose what it is most comfortable with, with at
// least the range required.

typedef unsigned rt_halcount8  rt_ucount8  ;
typedef   signed rt_halcount8  rt_count8   ;

typedef unsigned rt_halcount16 rt_ucount16 ;
typedef   signed rt_halcount16 rt_count16  ;

typedef unsigned rt_halcount32 rt_ucount32 ;
typedef   signed rt_halcount32 rt_count32  ;

typedef unsigned rt_halcount64 rt_ucount64 ;
typedef   signed rt_halcount64 rt_count64  ;

// -------------------------------------------------------------------------
// Define a type to be used for atomic accesses. This type is guaranteed
// to be read or written in a single uninterruptible operation. This type
// is at least a single byte.

typedef volatile unsigned rt_halatomic  rt_atomic;
typedef volatile unsigned rt_halatomic  RT_ATOMIC;

// -------------------------------------------------------------------------
// Define types for access plain, on-the-metal memory or devices.

typedef rt_uint32  RT_WORD;
typedef rt_uint8   RT_BYTE;
typedef rt_uint16  RT_WORD16;
typedef rt_uint32  RT_WORD32;
typedef rt_uint64  RT_WORD64;

//typedef	rt_uint32	rt_vector;

typedef rt_haladdress  RT_ADDRESS;
typedef rt_haladdrword RT_ADDRWORD;

// -------------------------------------------------------------------------
// Constructor ordering macros.  These are added as annotations to all
// static objects to order the constuctors appropriately.

#if defined(__cplusplus) && defined(__GNUC__)
# define RTBLD_ATTRIB_INIT_PRI( _pri_ ) __attribute__((init_priority(_pri_)))
#else
// FIXME: should maybe just bomb out if this is attempted anywhere else?
// Not sure
# define RTBLD_ATTRIB_INIT_PRI( _pri_ )
#endif
    
// The following will be removed eventually as it doesn't allow the use of
// e.g. pri+5 format
#define RT_INIT_PRIORITY( _pri_ ) RTBLD_ATTRIB_INIT_PRI( RT_INIT_##_pri_ )

#define RTBLD_ATTRIB_INIT_BEFORE( _pri_ ) RTBLD_ATTRIB_INIT_PRI(_pri_-100)
#define RTBLD_ATTRIB_INIT_AFTER( _pri_ )  RTBLD_ATTRIB_INIT_PRI(_pri_+100)

#define RT_INIT_HAL                    10000
#define RT_INIT_SCHEDULER              11000
#define RT_INIT_INTERRUPTS             12000
#define RT_INIT_DRIVERS                13000
#define RT_INIT_CLOCK                  14000
#define RT_INIT_IDLE_THREAD            15000
#define RT_INIT_THREADS                16000
#define RT_INIT_KERNEL                 40000
#define RT_INIT_IO                     49000
#define RT_INIT_LIBC                   50000
#define RT_INIT_COMPAT                 55000
#define RT_INIT_APPLICATION            60000
#define RT_INIT_PREDEFAULT             65534
#define RT_INIT_DEFAULT                65535

// -------------------------------------------------------------------------
// Label name macros. Some toolsets generate labels with initial
// underscores and others don't. RT_LABEL_NAME should be used on
// labels in C/C++ code that are defined in assembly code or linker
// scripts. RT_LABEL_DEFN is for use in assembly code and linker
// scripts where we need to manufacture labels that can be used from
// C/C++.
// These are default implementations that should work for most targets.
// They may be overridden in basetype.h if necessary.

#ifndef RT_LABEL_NAME

#define RT_LABEL_NAME(_name_) _name_

#endif

#ifndef RT_LABEL_DEFN

#define RT_LABEL_DEFN(_label) _label

#endif

// -------------------------------------------------------------------------
// COMPILER-SPECIFIC STUFF

#ifdef __GNUC__
// Force a 'C' routine to be called like a 'C++' contructor
# define RTBLD_ATTRIB_CONSTRUCTOR __attribute__((constructor))

// Define a compiler-specific rune for saying a function doesn't return
# define RTBLD_ATTRIB_NORET __attribute__((noreturn))

// How to define weak symbols - this is only relevant for ELF and a.out,
// but that won't be a problem for eCos
# define RTBLD_ATTRIB_WEAK __attribute__ ((weak))

// How to define alias to symbols. Just pass in the symbol itself, not
// the string name of the symbol
# define RTBLD_ATTRIB_ALIAS(__symbol__) \
        __attribute__ ((alias (#__symbol__)))

// This effectively does the reverse of the previous macro. It defines
// a name that the attributed variable or function will actually have
// in assembler.
#define __Str(x) #x
#define __Xstr(x) __Str(x)
# define RTBLD_ATTRIB_ASM_ALIAS(__symbol__) \
            __asm__ ( __Xstr( RT_LABEL_DEFN( __symbol__ ) ) )

// Shows that a function returns the same value when given the same args, but
// note this can't be used if there are pointer args
# define RTBLD_ATTRIB_CONST __attribute__((const))

// Assign a defined variable to a specific section
# define RTBLD_ATTRIB_SECTION(__sect__) __attribute__((section (__sect__)))

// Give a type or object explicit minimum alignment
# define RTBLD_ATTRIB_ALIGN(__align__) __attribute__((aligned(__align__)))

#else // non-GNU

# define RTBLD_ATTRIB_CONSTRUCTOR

# define RTBLD_ATTRIB_NORET
    // This intentionally gives an error only if we actually try to
    // use it.  #error would give an error if we simply can't.
// FIXME: Had to disarm the bomb - the RTBLD_ATTRIB_WEAK macro is now
//        (indirectly) used in host tools.
# define RTBLD_ATTRIB_WEAK /* !!!-- Attribute weak not defined --!!! */

# define RTBLD_ATTRIB_ALIAS(__x__) !!!-- Attribute alias not defined --!!!

# define RTBLD_ATTRIB_ASM_ALIAS(__symbol__) !!!-- Asm alias not defined --!!!

# define RTBLD_ATTRIB_CONST

# define RTBLD_ATTRIB_ALIGN(__align__) !!!-- Alignment alias not defined --!!!

#endif

// How to define weak aliases. Currently this is simply a mixture of the
// above

# define RTBLD_ATTRIB_WEAK_ALIAS(__symbol__) \
        RTBLD_ATTRIB_WEAK RTBLD_ATTRIB_ALIAS(__symbol__)

// -------------------------------------------------------------------------
// Various "flavours" of memory regions that can be described by the 
// Memory Layout Tool (MLT).

#define RTMEM_REGION_ATTR_R  0x01  // Region can be read
#define RTMEM_REGION_ATTR_W  0x02  // Region can be written

// -------------------------------------------------------------------------
#endif // RTONCE_INFRA_RT_TYPE_H multiple inclusion protection
// EOF rt_type.h
